structure ElectioneeringCop :> COPFUN =
struct
    open Message
    open Super
    open Graph

    exception Failure of string

    exception State of (pid * int) list * exn
               (* past election results, parent cop state *)

    fun register (cop : Super.cop) () : register_msg * state =
        let val ((pid, ptype), s) = #register cop ()
        in
            (* mod_electioneering!  woot!!11 *)
            ((pid ^ "#Electioneering", ptype), State ([], s))
        end

    fun world_skeleton (cop : Super.cop) (u, ws_msg, (State (_, s))) : state =
        let val s = #world_skeleton cop (u, ws_msg, s)
            val l = List.map (fn pid => (pid, 0)) (get_cops_as_list u)
                    (* alist mapping every cop to 0 *)
        in
            State (l, s)
        end

      | world_skeleton _ _ = raise Failure "bad state in electioneering-cop.sml:world_skeleton"

    fun inform cop (u, w_msg, State (l, s)) : inform_msg * state =
        let val (inf_msg, s) = #inform cop (u, w_msg, s)
        in
            (inf_msg, State (l, s))
        end
      
      | inform _ _ = raise Failure "bad state in electioneering-cop.sml:inform"

    fun plan cop (u, finf_msg, State (l, s)) =
        let val (p_msg, s) = #plan cop (u, finf_msg, s)
        in
            (p_msg, State (l, s))
        end
      
      | plan _ _ = raise Failure "bad state in electioneering-cop.sml:plan"

    fun vote cop (u, fplan_msg, State (l, s)) =
        let val (_, s) = #vote cop (u, fplan_msg, s)
            val mypid = get_name u
            (* vote for the least vote getter first, greatest vote getter last,
               except always vote for me zeroth *)
            val l_without_me = List.filter (fn (pid, _) => pid <> mypid) l
            val l_wo_me_sort = ListUtil.sort (fn ((_, vs1), (_, vs2)) =>
                                                Int.compare (vs1, vs2))

                                             l_without_me
            val v_msg = mypid :: (map (fn (pid, _) => pid) l_wo_me_sort)
        in
            (v_msg, State (l, s))
        end

      | vote _ _ = raise Failure "bad state in electioneering-cop.sml:vote"

    fun move cop (u, vres_msg, State (l, s)) =
        let val (mv_msg, s) = #move cop (u, vres_msg, s)
            val l =
                case vres_msg of 
                    NONE => l
                    (* update the Alist if somebody's a winner *)
                  | SOME pid => ListUtil.Alist.modify op= l pid (fn x => x + 1)
        in
            (mv_msg, State (l, s))
        end
    
      | move _ _ = raise Failure "bad state in electioneering-cop.sml:move"

    fun extend cop =
        { register = register cop,
          world_skeleton = world_skeleton cop,
          inform = inform cop,
          plan = plan cop,
          vote = vote cop,
          move = move cop }
          
end
