structure Friday :> COP = 
struct

  structure T = Types
  structure S = Super
  structure G = Graph

  exception ST of exn list

  val cops =
      [
       ("McGruff", McGruff.new nil),
       ("McGruffElect", ElectioneeringCop.extend (McGruff.new nil)),
       ("NoOpCop", NoOpCop.new nil),
       ("RandomCop", RandomCop.new nil),
       ("RandomElectCop", ElectioneeringCop.extend (RandomCop.new nil))
       ]

  val REGISTER = Params.param "McGruff" (SOME ("-register", 
                                               "Which cop should register?"))
                              "friday-register"
  val INFORM = Params.param "McGruff" (SOME ("-inform", 
                                             "Which cop should inform?"))
                            "friday-inform"
  val PLAN = Params.param "McGruff" (SOME ("-plan", 
                                           "Which cop should plan?"))
                          "friday-plan"
  val VOTE = Params.param "McGruff" (SOME ("-vote", 
                                           "Which cop should vote?"))
                          "friday-vote"
  val MOVE = Params.param "RandomCop" (SOME ("-move", 
                                             "Which cop should move?"))
                          "friday-move"

  val register_weights = ref nil : int list ref
  val inform_weights = ref nil : int list ref
  val plan_weights = ref nil : int list ref
  val vote_weights = ref nil : int list ref
  val move_weights = ref nil : int list ref

  exception Impossible

  fun compute_weights _ =
      let fun compute_weight name =
              List.map (fn (name', _) => if name = name' then 10 else 0) cops
          val _ = register_weights := compute_weight (!REGISTER)
          val _ = inform_weights := compute_weight (!INFORM)
          val _ = plan_weights := compute_weight (!PLAN)
          val _ = vote_weights := compute_weight (!VOTE)
          val _ = move_weights := compute_weight (!MOVE)
      in 
        ()
      end

  fun compute_scores props V =
      List.foldl (fn ((p, wt), V) => 
                     List.map (fn (p', v) => 
                                  (p', if p = p' then v + wt else v))
                              V) 
                 V props

  fun find_max (x, v) V = 
      List.foldl (fn ((x, v), (best_x, best_v)) => 
                     if v > best_v then (x, v) else (best_x, best_v)) 
                 (x, v) V

  fun register () : S.register_msg * S.state = 
      let val _ = compute_weights ()
          fun do_register (_, cop) =
              (#register cop) ()
          val results = List.map do_register cops

          val pts = List.map (fn (((_, pt), _), wt) => (pt, wt)) 
                             (ListPair.zip (results, !register_weights))

          val V = [(T.CopFoot, 0), (T.CopCar, 0)]
          val V = compute_scores pts V

          val (pt, _) = find_max (T.CopFoot, 0) V
          val states = List.map (fn (_, st) => st) results
      in
        (("Friday", pt), ST states)
      end

  fun world_skeleton (u, ws_msg, ST states) : S.state = 
      let fun do_world_skeleton ((_, cop), s) =
              (#world_skeleton cop) (u, ws_msg, s)
          val states = List.map do_world_skeleton (ListPair.zip (cops, states))
      in
        (ST states)
      end
    | world_skeleton _ = raise Impossible

  fun inform (u, w_msg, ST states) : S.inform_msg * S.state = 
      let fun do_inform ((_, cop), s) = (#inform cop) (u, w_msg, s)
          val results = List.map do_inform (ListPair.zip (cops, states))

          val i_msgs = List.map (fn ((i_msg, _), wt) => (i_msg, wt)) 
                                (ListPair.zip (results, !inform_weights))

          val (i_msg, _) = find_max (nil, 0) i_msgs
          val states = List.map (fn (_, st) => st) results
      in
        (i_msg, ST states)
      end
    | inform _ = raise Impossible

  fun plan (u, fi_msg, ST states) : S.plan_msg * S.state =
      let fun do_plan ((_, cop), s) = (#plan cop) (u, fi_msg, s)
          val results = List.map do_plan (ListPair.zip (cops, states))

          val p_msgs = List.map (fn ((p_msg, _), wt) => (p_msg, wt)) 
                                (ListPair.zip (results, !plan_weights))

          val (p_msg, _) = find_max (nil, 0) p_msgs
          val states = List.map (fn (_, st) => st) results
      in
        (p_msg, ST states)
      end
    | plan _ = raise Impossible

  fun vote (u, fp_msg, ST states) : S.vote_msg * S.state =
      let fun do_vote ((_, cop), s) = (#vote cop) (u, fp_msg, s)
          val results = List.map do_vote (ListPair.zip (cops, states))

          val v_msgs = List.map (fn ((v_msg, _), wt) => (v_msg, wt)) 
                                (ListPair.zip (results, !vote_weights))

          val (v_msg, _) = find_max (nil, 0) v_msgs
          val states = List.map (fn (_, st) => st) results
      in
        (v_msg, ST states)
      end
    | vote _ = raise Impossible

  fun move (u, vr_msg, ST states) : S.move_msg * S.state =
      let fun do_move ((_, cop), s) = (#move cop) (u, vr_msg, s)
          val results = List.map do_move (ListPair.zip (cops, states))
          val m_msgs = List.map (fn ((m_msg, _), wt) => (m_msg, wt)) 
                                (ListPair.zip (results, !move_weights))
          (* XXX consolidate? *)
          val (m_msg, _) = find_max (S.get_my_loc u, 0) m_msgs
          val states = List.map (fn (_, st) => st) results
      in
        (m_msg, ST states)
      end
    | move _ = raise Impossible

  fun new _ = {register=register, world_skeleton=world_skeleton,
               inform=inform, plan=plan, vote=vote, move=move}

end
