structure HandshakingCop :> COP =
struct

  structure T = Types
  structure S = Super
  structure G = Graph

  exception Impossible

  exception ST of T.pid list

  val SECRET = "CMU h4ck3rs rul3! y0u 4r3 0wn3d!!11"
  val OFFSET = 100
  val ANNOUNCE_TIME = 1

  fun encode (k, d) = (Char.chr ((Char.ord k) + (Char.ord d) - 30))
  fun decode (k, d) = (Char.chr ((Char.ord d) - (Char.ord k) + 30))

  (* Given a sequence of key characters and data characters, apply the given
   * function to each pair of values, recycling key values as necessary.  *)
  fun mangle_list f (ks, ds)
    = let fun mangle (k::ks', d::ds') = (f (k, d))::(mangle (ks', ds'))
            | mangle (_, nil) = nil
            | mangle (nil, ds) = mangle (ks, ds)
      in
        mangle (ks, ds)
      end

  fun register () : S.register_msg * S.state = 
      (("HandshakingCop", T.CopFoot), ST nil)

  fun world_skeleton (u, msg, s) : S.state = s

  fun inform (u, _, s) : S.inform_msg * S.state = 
      if S.get_time u = ANNOUNCE_TIME then
        let val pd = S.get_name u
            (* For this encoding to work the length of the secret (plus the
             * offset) must be less than the number of worlds, and there must 
             * be enough nodes to encode the ASCII character set.  We use
             * certainty of 0 so that we are not (really) lying. *)
            fun to_msg (c, i) = (pd, Char.ord c, T.CopFoot, i + OFFSET, 0)
            
            val cs = mangle_list encode (String.explode pd, 
                                         String.explode SECRET)
            val msg = ListUtil.mapi to_msg cs
            val _ = Spew.spew "Sending secret message..."
        in 
          (msg, s)
        end
      else (nil, s)

  fun plan (u, fi_msg, s) : S.plan_msg * S.state =
      if S.get_time u = ANNOUNCE_TIME then
        let fun is_friend (pd, i_msg) = 
                (* First filter out parts of the msg we don't care about. *)
                let
                  val i_msg = List.filter (fn (pd', _, pt, t, cer) =>
                                                  pd = pd' andalso
                                                  pt = T.CopFoot andalso
                                                  t >= OFFSET andalso
                                                  cer = 0) i_msg
                  val i_msg = List.map (fn (_, nd, _, t, _) => (nd, t)) i_msg

                  (* Sort by "t" just in case the server reordered . *)
                  val i_msg = ListUtil.sort (fn ((_, t), (_, t')) => 
                                                  Int.compare (t, t')) i_msg
                  val cs = (List.map (Char.chr o #1) i_msg)
                in
                  if (String.implode (mangle_list decode 
                                                  (String.explode pd, cs)))
                     = SECRET
                  then SOME pd
                  else NONE
                end
            val friends = List.mapPartial is_friend fi_msg
            val s = List.foldl (fn (pd, s) => s ^ pd ^ " ") "" friends
            val _ = Spew.spew ("Found friends: " ^ s)
        in
          (nil, ST friends)
        end
      else 
        (nil, s)
        
  fun vote (u, _, s) : S.vote_msg * S.state = (S.get_cops_as_list u, s)

  fun move (u, _, s) : S.move_msg * S.state
    = let val (nd, pt) = S.get_my_loc u
      in
        ((nd, pt), s)
      end

  fun new _ = {register=register, world_skeleton=world_skeleton,
               inform=inform, plan=plan, vote=vote, move=move}

end
