structure McGruff :> COP =
struct

  structure T = Types
  structure S = Super
  structure G = Graph

  exception Impossible

  exception ST

  fun register () = (("McGruff", T.CopFoot), ST)

  fun world_skeleton (u, msg, s) = s

  fun inform (u, w_msg, s) = 
      let val (nd, pt) = S.get_my_loc u
          val smell = S.get_smell u
          val dist = case smell
                      of SOME (d) => d
                       | NONE => case pt of T.CopFoot => 2
                                          | T.CopCar => 1
                                          | _ => raise Impossible
          val reachable = G.reachable (S.get_graph u) G.Forward (G.Only pt)
                                      [nd] dist
          val reachable = List.mapPartial (fn nd' => if nd = nd' 
                                                     then NONE else SOME nd')
                                          reachable
          val len = List.length reachable
          val robber = S.get_robber u
          val now = S.get_time u
      in
        case smell
         of SOME (d) => (List.map (fn nd => (robber, nd, T.Robber, now, 
                                             100 div (len - 1)))
                                  reachable, s)
          | NONE => (List.map (fn nd => (robber, nd, T.Robber, now, ~100))
                              (nd :: reachable), s)
      end

  fun plan (u, msg, s) = (nil, s)

  fun vote (u, msg, s) = (S.get_cops_as_list u, s)

  fun move (u, SOME(pid), s) =
      let val plans = S.get_plans_by_planner (u, pid)
          val name = S.get_name u
          val wd = S.get_time u
          val plan = List.find (fn (pid', _, _, wd') 
                                   => name = pid' andalso wd = wd')
                               plans
      in
        (* XXX should mcgruff check the validity of the proposed move? *)
        case plan of SOME (_, nd, pt, _) => ((nd, pt), s)
                   | NONE => (S.get_player_loc (u, name), s)
      end
    (* If there was no winner, then don't move. *)
    | move (u, NONE, s) = (S.get_my_loc u, s)

  fun new _ = {register=register, world_skeleton=world_skeleton,
               inform=inform, plan=plan, vote=vote, move=move}

end
