structure RandomCop :> COP =
struct

  structure T = Types
  structure S = Super
  structure G = Graph

  exception Impossible

  exception ST

  val SEED = Params.param "0" (SOME ("-seed", "RandomCop seed")) 
                              "randomcop-seed"

  fun register () = (("RandomCop", T.CopFoot), ST)

  fun world_skeleton (u, msg : S.world_skeleton_msg, s) =
        let val seed = case Params.asint 0 SEED
                        of 0 => (case MLton.Random.seed () of
                                   NONE => 0w0
                                 | SOME x => x)
                         | x => Word.fromInt x
            val () = MLton.Random.srand seed
            val () = Spew.spew ("Cop " ^ #name msg ^ ": random seed = "
                                ^ Word.toString seed)
         in
             s
         end

  (* XXX Cut-and-paste from McGruff! *)
  fun inform (u, w_msg, s) = 
      let val (nd, pt) = S.get_my_loc u
          val smell = S.get_smell u
          val dist = case smell
                      of SOME (d) => d
                       | NONE => case pt of T.CopFoot => 2
                                          | T.CopCar => 1
                                          | _ => raise Impossible
          val reachable = G.reachable (S.get_graph u) G.Forward (G.Only pt)
                                      [nd] dist
          val reachable = List.mapPartial (fn nd' => if nd = nd' 
                                                     then NONE else SOME nd')
                                          reachable
          val len = List.length reachable
          val robber = S.get_robber u
          val now = S.get_time u
      in
        case smell
         of SOME (d) => (List.map (fn nd => (robber, nd, T.Robber, now, 
                                             100 div (len - 1)))
                                  reachable, s)
          | NONE => (List.map (fn nd => (robber, nd, T.Robber, now, ~100))
                              (nd :: reachable), s)
      end

  fun randRange (lo, hi) =
      let val w = MLton.Random.rand ()
          open Word
          val word = Word.fromInt
          val int = Word.toInt
      in
          (int) (w mod (word hi - word lo + word 1) + word lo)
      end

  fun permute lst =
      let val a = Array.fromList lst
	  val len = Array.length a

	  fun to_list a = Array.foldr (op ::) [] a
		    
	  fun swap (m,n,a) =
	      let val a_m = Array.sub(a, m)
		  val a_n = Array.sub(a, n)
	      in
		  Array.update(a, m, a_n);
		  Array.update(a, n, a_m)
	      end

	  fun loop i =
	      if i < len then
		  (swap(i, randRange(0, len-1), a);
		   loop (i+1))
	      else
		  to_list a
      in
	  loop 0
      end


  fun plan (u, msg, s) = (nil, s)

  fun vote (u, msg, s) = (permute(S.get_cops_as_list u), s)

  fun move (u, msg, s) =
      let val (nd, myptype) = S.get_my_loc u
          val reachable = G.adjacent (S.get_graph u) G.Forward
                                      (G.Any myptype) nd
          val len = List.length reachable
          val i = randRange (0, len - 1)
          val (nd, pt) = List.nth (reachable, i)
      in
        ({moves = [(nd, pt, S.get_name u)],
          offer = S.StraightArrow,
          accuse = nil}, s)
      end

  fun r_inform (u, _, s) : S.inform_msg * S.state = (nil, s)

  fun r_plan (u, _, s) : S.plan_msg * S.state = (nil, s)

  fun r_vote (u, fp_msg, s) : S.vote_msg * S.state =
      (List.map (fn (bd, _) => bd) fp_msg, s)

  fun r_move (u, _, s) : S.dirty_cop_move_msg * S.state
    = let val (nd, pt) = S.get_my_loc u
      in
        ([(nd, pt, S.get_name u)], s)
      end

  fun r_results (u, _, s) : S.state = s

  fun dirtify (u, s) = ({inform=inform, plan=plan, vote=vote, move=r_move,
                         r_inform=r_inform, r_plan=r_plan, r_vote=r_vote,
                         r_results=r_results}, s)

  fun new _ = {register=register, world_skeleton=world_skeleton,
               dirtify=dirtify, inform=inform, plan=plan, vote=vote, move=move}

end
