(* I am Pessimus Prime!
   I am Pessimus Prime!
   I am Pessimus Prime!
   I am Pessimus Prime! *)


structure EvaluateRobber' =
struct

    open Transform

    (* XXX these constants need tuning!!
       I don't even have a good idea for the relative
       magnitudes of these arbitrary constants *)

    val UNIVERSAL_RADIANCE_CONSTANT = 0.4

    val CASH_WEIGHT = 5
    val TEMPERATURE_WEIGHT = 100
    val RADIANT_WEIGHT = 4

    (* XXX should look at time too, and do something different when
       time = endgame. (for instance, we should only worry about local
       information like heat at the current node and temperature, and
       banks right where we're standing) this can be generalized to when
       we are at time endgame-n for n<diameter_of_graph; no need to
       consider anything outside that diameter. *)

    (* since I haven't settled on a rep for cops, this only
       evaluates the loot we've collected and our proximity
       to money. *)
    fun evaluate (GS{ skel, time, loot, bank_vals, 
                      evidence, temp, 
                      cops = {foot, car} : RobberHeatTransform.cops_asheat, 
                      robber : exactrobber, transfers }) =               
        let
            (* the worth of a bank is its
               cash amount,
               divided by the square of the distance to it. *)
            fun bank_worth (node, value) =
                let val dist = real (Graph.distance skel
                                        (Graph.Only Types.Robber)
                                     robber node)
                in (real value) / (dist + 1.0)
                end

            (* now penalize for being near heat *)

            val radiant_heat = ref 0.001

            (* value of the heat in one spot *)
            fun value NONE = 0
              | value (SOME (x, _)) = x

            val _ = 
                Util.for 0 (Vector.length foot - 1)
                (fn n =>
                 let 
                     val fd = 1 + Graph.distance skel (Graph.Any Types.CopFoot) n robber 
                     val cd = 1 + Graph.distance skel (Graph.Any Types.CopCar)  n robber 
                     val fh = value (Vector.sub(foot, n))
                     val ch = value (Vector.sub(car, n))
                 in
                     (* inverse square law with distance.. *)
                     (* XXX probably should use 1/(avg_degree^dist) *)
                     radiant_heat := !radiant_heat +
                        real fh / (UNIVERSAL_RADIANCE_CONSTANT * real fd * real fd) +
                        real ch / (UNIVERSAL_RADIANCE_CONSTANT * real cd * real cd)
                 end)


            (* could also easily figure in discounted factors
               for banks that will soon receive transfers *)
            val best_bank = ListUtil.max Real.compare
                            (map bank_worth bank_vals)
            val potential = best_bank + (real loot)
            val danger = real (temp * TEMPERATURE_WEIGHT) +
                         (!radiant_heat) * (real RADIANT_WEIGHT)
(*
            val _ = Spew.spew ("potential " ^ Real.toString potential ^"\n")
            val _ = Spew.spew ("temp " ^ Int.toString temp ^"\n")
            val _ = Spew.spew ("heat " ^ Real.toString (!radiant_heat))
*)
        in
            potential / danger
        end

end

structure PessimusPrime :> ROBBER =
  struct

    structure T = Types
    structure S = Super
    structure G = Graph

    open Transform

    exception PessimusPrime of string

    val HEATFACTOR = 4
        
    (* number of robber moves to search deep
       (minus one)
       
       this should be tuned so that we can search
       as deep as possible within a turn. *)
    val DEPTH = 3

    exception ST of
       { graph : Graph.graph,
         transfers : (Graph.nid * int) list,
         evidence : (Graph.nid * int) list,
         oldheat : RobberHeatTransform.cops_asheat,
         lastworld : S.world_msg option }

    (*
    val RADIUS = Params.param "1" (SOME ("-radius", 
                                         "Francois scoring radius")) 
                              "francois-radius"
                              *)

    exception NullPointer (* ;) *)
    (* start with bad exn, since we can't initialize yet *)
    fun register () = (("PessimusPrime", T.Robber), NullPointer)

    fun world_skeleton (u, ws_msg : S.world_skeleton_msg, _) = 

        let val g = #graph ws_msg
            val cold = Vector.tabulate(Graph.length g,
                                       fn _ => NONE)
        in
            ST { graph = g, transfers = nil, evidence = nil,
                 oldheat = { foot = cold,
                             car = cold }, lastworld = NONE }
        end
        


    (* evalrec d st move
       
       evaluate a move for a gamestate (on robber's turn) recursively. 
       larger results are better. *)
    fun evalrec 0 (st as GS{time,...}) move =
        let
            (* val _ = Spew.spew ("(1)Time is: " ^ Int.toString time ^ "\n") *)
            val stres = TransformByRobber.transform st move
        in
            EvaluateRobber'.evaluate st
        end
      | evalrec n st move =
        let
            (* XXX boundary condition for game over *)
            val stres = TransformByRobber.transform st move
            val stres = RobberHeatTransform.transform stres

            fun next m = evalrec (n - 1) stres m

            val opts = TransformByRobber.enumerate stres
            val scores = map next opts
        in
            (* we'll get to choose, so just use best score
               (maximax) *)
            ListUtil.max Real.compare scores
        end


    val evalrec = fn x => (evalrec x handle e => raise e)
        
    fun move (u, offered_cops_msg, 
              ST { graph, 
                   transfers,
                   evidence,
                   oldheat,
                   lastworld = 
                   lastworld as
                      SOME {bank_vals, 
                            (* XXX more good stuff here now *)
                            ...} } ) =
        let 
            val (me, _) = S.get_my_loc u


            (* null hypothesis: robber could be anywhere *)
            val flatslime = 
                Vector.tabulate(Graph.length graph, 
                                fn _ => RobberHeatTransform.SLIMEFACTOR)

            val cops = 
                List.mapPartial (fn (_, n, T.CopFoot) => SOME (n, false)
                                 |  (_, n, T.CopCar) => SOME (n, true)
                                 | _ => NONE) (S.get_player_locs u)

            val num = Graph.length graph

            fun oldslime car i =
                let val a = if car 
                            then #car oldheat
                            else #foot oldheat
                in
                    case Vector.sub(a, i) of
                        NONE => flatslime
                      | SOME (_, v) => v
                end

            val heat =
                {foot = 
                 Vector.tabulate
                 (num,
                  fn i =>
                  let val ncops =
                      length (List.filter (fn (x, false) =>
                                           x = i | _ => false) cops)
                      * num * HEATFACTOR
                  in
                      if ncops > 0
                      then SOME (ncops, oldslime false i)
                      else NONE
                  end),

                 car = 
                 Vector.tabulate
                 (num,
                  fn i =>
                  let val ncops =
                      length (List.filter (fn (x, true) => x = i
                                           | _ => false) cops)
                      * num * HEATFACTOR
                  in
                      if ncops > 0
                      then SOME (ncops, oldslime true i)
                      else NONE
                  end) }

            (* initialize gamestate. *)
            val start 
                : (RobberHeatTransform.cops_asheat, 
                   Transform.exactrobber) Transform.gamestate
                =
                GS { skel = graph,
                     time = S.get_time u,
                     loot = S.get_loot u,
                     bank_vals = bank_vals,
                     temp = 0, (* safe for now *)
                     cops = heat,
                     robber = me,
                     (* we have to remember these ourselves,
                        since the world does not show them *)
                     evidence = evidence,
                     transfers = transfers }


            (* search *)
            fun try m = (evalrec DEPTH start m, m)
            val opts = TransformByRobber.enumerate start
            val withscores = map try opts

            (*
            val _ = Spew.spew (Int.toString (length withscores) ^ " moves:\n")
            val _ = 
                app (fn (sc, move) =>
                     Spew.spew ("  " ^ 
                                Graph.nid_to_name graph move ^ " score " ^
                                Int.toString sc ^ "\n")) withscores
                *)

            val (_, bestmove) = 
                (* always at least one, because of self loop *)
                ListUtil.max (ListUtil.byfirst Real.compare) withscores

(*            val _ = Spew.spew ("Time is: " ^ Int.toString (S.get_time u) ^ "\n") *)

            (* PERF this is done somewhere above *)
            val newstate = TransformByRobber.transform start bestmove
            val newstate = RobberHeatTransform.transform newstate

(*             val _ = Spew.spew "newstate..\n" *)

            val GS{transfers, evidence, cops, ...} = newstate
            fun stupidpush () =
                let val victim = hd (S.get_cops_as_list u)
                    val (nd, pt) = S.get_player_loc (u, victim)
                    val moves = G.adjacent graph G.Forward (G.Only pt) nd
                    val somend = #1 (hd moves)
                in
                   S.Push (victim, somend)
                end

            val innocent_world = S.get_dirty_cops u = [] andalso
                                 S.get_controlling u = []
            val bribe = case offered_cops_msg of
                             S.NotTelling => S.BribeNone
                           | S.Offered cops =>
                             if cops = [] andalso innocent_world
                             then stupidpush ()
                             else S.BribeChoose cops
        in
            ({ move = bestmove, result = bribe },
             ST { graph = graph,
                  transfers = transfers,
                  evidence = evidence,
                  oldheat = cops,
                  lastworld = NONE })
        end
      | move _ = raise PessimusPrime "driver error: wrong state (no lastworld?)"

    val move = fn x => (move x handle e => raise e)


    (* XXX no-ops! replace these with the good stuff. *)
    (* but here we have to remember the world message for later.. *)

    (* IMO inform is totally useless, since both the robbers and the cops
       know exactly where everyone is. We don't need to or want to
       trick anyone (maybe at the end of the round we might want our
       helper bots to get caught so that they don't share our loot?),
       so what is the point? *)
    fun inform (u, wm, ST {graph, transfers, evidence, oldheat, ... }) 
        : S.inform_msg * S.state = 
        (nil, 
         ST { graph = graph,
              transfers = transfers, 
              evidence = evidence, 
              oldheat = oldheat,
              lastworld = SOME wm })
      | inform (_, _, _) = raise PessimusPrime "wrong state for inform"

    (* XXX propose reasonable plan *)
    fun plan (u, _, s) : S.plan_msg * S.state = (nil, s)

    fun vote (u, fp_msg, s) : S.vote_msg * S.state = 
        (List.map (fn (bid, _) => bid) fp_msg, s)
        
    fun bribe (u, vres_msg, s) : S.bribe_msg * S.state =
        if (S.get_loot u >= 50) then (S.NoBribe, s) else (S.NoBribe, s)

    fun new _ : Super.robber =
        { register = register,
          world_skeleton = world_skeleton,
          move = move,
          inform = inform,
          plan = plan,
          vote = vote,
          bribe = bribe }
  end
