structure RandomRobber :> ROBBER =
  struct

    structure T = Types
    structure S = Super
    structure G = Graph

    exception S

    fun register () = (("RandomRobberX", Message.Robber), S)

    fun world_skeleton (u, ws_msg : S.world_skeleton_msg, s) =
        let val seed =
                case MLton.Random.seed () of
                    NONE => 0w0
                  | SOME x => x
            val () = MLton.Random.srand seed
            val () = Spew.spew ("Robber " ^ #name ws_msg ^ ": random seed = "
                                ^ Word.toString seed)
         in
             s
         end

    fun inform (u, _, s) : S.inform_msg * S.state = (nil, s)

    fun plan (u, _, s) : S.plan_msg * S.state = (nil, s)

    fun vote (u, fp_msg, s) : S.vote_msg * S.state = 
        (List.map (fn (bd, _) => bd) fp_msg, s)

    fun bribe (u, _, s) : S.bribe_msg * S.state = (S.NoBribe, s)

    fun move (u, _ : S.offered_cops_msg, s) : S.robber_move_msg * S.state =
        let fun randRange (lo, hi) =
                let val w = MLton.Random.rand ()
                    open Word
                    val word = Word.fromInt
                    val int = Word.toInt
                in
                    (int) (w mod (word hi - word lo + word 1) + word lo)
                end
            val (nd, myptype) = S.get_my_loc u
            val reachable = G.reachable (S.get_graph u) G.Forward
                                        (G.Only myptype) [nd] 1
            val len = length reachable
            val i = randRange (0, len - 1)
            val nd = List.nth (reachable, i)
        in
          ({move = nd, result = S.BribeNone}, s)
        end

    fun new _ : S.robber =
        { register = register,
          world_skeleton = world_skeleton,
          inform = inform, plan = plan, vote = vote, bribe = bribe,          
          move = move }
  end
