#!/usr/bin/perl -w

use 5.008;
use strict;
use bytes;
use List::Util 'shuffle';
use Time::HiRes qw( usleep );
use File::Copy;
use File::Temp qw( tempfile );

my @chars = ('|', '/', '-', '\\');
my $char_index = 0;

my $last_get_time = 0;
my @last_items;

$| = 1;

while (1) {
  if (-f "./dfa2000/results/SLAVES_EXIT_NOW") {
    exit;
  }

  # get bots
  my @result = get_bots();
  my @cops = @{$result[0]};
  my @robbers = @{$result[1]};

  unless (@cops) {
    print "No cops!\n";
    sleep 30;
    next;
  }

  unless (@robbers) {
    print "No robbers!\n";
    sleep 30;
    next;
  }


  # generate random run
  @cops = shuffle(@cops);
  @robbers = shuffle(@robbers);

  my $robber = $robbers[0];
  my @cops5 = @cops[0..4];
  @cops = sort {$a->[0] cmp $b->[0]} @cops5;

  my $r_name = $robber->[0];
  my $r_invocation = $robber->[1];

  my @c_names = map {$_->[0]} @cops;
  my @c_invocations = map {$_->[1]} @cops;


  # check it
  my $output_name = "./dfa2000/results/" . join("-", @c_names) . "-$r_name.out";

  if (-s $output_name) {
    print "\rSearching for combinations to run $chars[$char_index]";
    $char_index = ($char_index + 1) % (scalar @chars);

    usleep(100000);
    next;
  }


  # here we go
  #print "Going to save result to $output_name...\n";

  print "\nR: $r_name, C: [" . join(",", @c_names) . "]\n";

  my ($fh, $tempfile) = 
    tempfile( "dfa2000slaveXXXXXX", 
	      PREFIX => "slave", 
	      DIR => "./dfa2000/results", SUFFIX => ".tmp" );

  my @cmd = ("/bin/bash", "-c",
	     "./bdk/tw-server \"\$@\" 2>$tempfile",
	     "server",
	     #"-g",
	     "-a", $c_invocations[0] . " -botname 1cop",
	     "-b", $c_invocations[1] . " -botname 2cop",
	     "-c", $c_invocations[2] . " -botname 3cop",
	     "-d", $c_invocations[3] . " -botname 4cop",
	     "-e", $c_invocations[4] . " -botname 5cop",
	     "-r", $r_invocation . " -botname 6rob",);
  #print "Running '" . join ("^", @cmd) . "'\n";

  open (SERVER, "-|", @cmd) or die "Cannot start server: $!";

  open OUTPUT, ">$output_name" or die "Can't open $output_name: $!";
  my @bad_output;
  my $scores_next_line = 0;
  while(<SERVER>) {
    #print "reading line from server\n";
    if ($scores_next_line) {
      print OUTPUT "scores $_";
      $scores_next_line = 0;
    } elsif (/^scores/) {
      $scores_next_line = 1;
    } else {
      print "**** bad output from server: $_\n";
      push @bad_output, $_;
    }
  }
  close OUTPUT;
  close SERVER;

  if (@bad_output) {
    open ERR, ">>$tempfile" or die "Can't open $tempfile: $!";
    foreach (@bad_output) {
      print ERR "$_\n";
    }
    close ERR;
  }

  move($tempfile, "$output_name.err");
}


sub get_bots {
  if (time - $last_get_time < 30 && @last_items) {
    return @last_items;
  }


  # build handled in java code now

  #print "\nUpdating... ";
  #open CVS, "-|", "cvs", "up" or die "Can't cvs up: $!";
  #while($_ = <CVS>) {
  #  chomp;
  #  next unless /^[PU]/;
  #  print "$_\n";
  #}

  #system("make");

  my %cop_names;
  my %robber_names;

  print "\nScanning botlist...";
  open BOTLIST, "./dfa2000/bots-list.txt" or die "Can't open bots list: $!";
  my @cops;
  my @robbers;
  while($_ = <BOTLIST>) {
    chomp;

    # skip comment
    next if (/^#/);

    # skip blanks
    next if (/^\s*$/);

    # read
    unless (/^(C|R)\s+(\S+)\s+(\S+)([^#]+)/) {
      next;
    }

    my $type = $1;
    my $name = $2;
    my $invocation = $3 . " \@MLton ram-slop .12 -- " . $4;

    if ($type eq "C") {
      if (defined $cop_names{$name}) {
	print "*** Warning! Ignoring duplicate cop: $name\n";
      } else {
	$cop_names{$name} = 1;
	push @cops, [$name, $invocation];
      }
    } elsif ($type eq "R") {
      if (defined $robber_names{$name}) {
	print "*** Warning! Ignoring duplicate robber: $name\n";
      } else {
	$robber_names{$name} = 1;
	push @robbers, [$name, $invocation];
      }
    } else {
      print "Unknown type: $type\n";
    }
  }

  while (@cops && (@cops < 5)) {
    @cops = shuffle(@cops);
    push @cops, $cops[0];
  }


  my $num_r = @robbers;
  my $num_c = @cops;

  print " $num_c cops, $num_r robbers ";

  my $combinations = $num_r * choose($num_c, 5);
  print "($combinations combinations total)\n";

  $last_get_time = time;
  @last_items = (\@cops, \@robbers);
  return @last_items;
}

sub choose {
  my ($n, $k) = @_;

  if ($n == 0) {
    return 0;
  }

  #print "$n choose $k\n";

  return fact($n) / (fact($k) * fact($n - $k));
}

sub fact {
  my $n = $_[0];

  if ($n == 1 || $n == 0) {
    return 1;
  } elsif ($n < 0) {
    return 0;
  } else {
    return $n * fact($n - 1);
  }
}
