
(* Part of the "Transform" series; this evaluates a gamestate from
   the robber's perspective. Positive results correspond to good
   boards, negative ones to bad boards. 

   Note: this does not do search! This is what we use to evaluate
   board positions at the leaves of search. *)

structure EvaluateRobber =
struct

    open Transform

    (* XXX these constants need tuning!!
       I don't even have a good idea for the relative
       magnitudes of these arbitrary constants *)

    val UNIVERSAL_RADIANCE_CONSTANT = 0.4

    val CASH_WEIGHT = 5
    val TEMPERATURE_WEIGHT = 100
    val RADIANT_WEIGHT = 4

    (* XXX should look at time too, and do something different when
       time = endgame. (for instance, we should only worry about local
       information like heat at the current node and temperature, and
       banks right where we're standing) this can be generalized to when
       we are at time endgame-n for n<diameter_of_graph; no need to
       consider anything outside that diameter. *)

    (* since I haven't settled on a rep for cops, this only
       evaluates the loot we've collected and our proximity
       to money. *)
    fun evaluate (GS{ skel, time, loot, bank_vals, 
                      evidence, temp, 
                      cops = {foot, car} : RobberHeatTransform.cops_asheat, 
                      robber : exactrobber, transfers }) =               
        let
            (* the worth of a bank is its
               cash amount,
               divided by the square of the distance to it. *)
            fun bank_worth (node, value) =
                   value
                    div
               (Graph.distance skel (Graph.Only Types.Robber) robber node + 1)

            (* now penalize for being near heat *)

            val radiant_heat = ref 0.001

            (* value of the heat in one spot *)
            fun value NONE = 0
              | value (SOME (x, _)) = x

            val _ = 
                Util.for 0 (Vector.length foot - 1)
                (fn n =>
                 let 
                     val fd = 1 + Graph.distance skel (Graph.Any Types.CopFoot) n robber 
                     val cd = 1 + Graph.distance skel (Graph.Any Types.CopCar)  n robber 
                     val fh = value (Vector.sub(foot, n))
                     val ch = value (Vector.sub(car, n))
                 in
                     (* inverse square law with distance.. *)
                     (* XXX probably should use 1/(avg_degree^dist) *)
                     radiant_heat := !radiant_heat +
                        real fh / (UNIVERSAL_RADIANCE_CONSTANT * real fd * real fd) +
                        real ch / (UNIVERSAL_RADIANCE_CONSTANT * real cd * real cd)
                 end)


            (* val potential = foldl op+ loot (map bank_worth bank_vals) *)
        in
            (* could also easily figure in discounted factors
               for banks that will soon receive transfers *)

            (* XXX this should be updated according to noam's suggestion.
               the idea is that we should be more conservative when we have
               more cash in hand already, so something like
               
               value = (1-P(death))*(loot-current_loot + probable_future_loot(time))
                      - P(death) * current_loot
               
               *)

            (* cash value *)
            real ((foldl op+ (loot * CASH_WEIGHT) (map bank_worth bank_vals))

            (* discount by danger *)

                 (* might already be dead *)
                 - (TEMPERATURE_WEIGHT * temp)
                 (* penalize for being near robbers *)
                                        (* XXX cops? *)
                 - (Real.trunc (!radiant_heat * (real RADIANT_WEIGHT))))

            (* (real potential) / (!radiant_heat) *)
        end

end
