signature GRAPH = 
sig

    exception Graph of string
    
    type nid = Types.nid
    type node = Types.node
    type edge_tag = Types.edge_tag
    type edge = Types.edge
    type ptype = Types.ptype
        
    datatype direction = Forward | Backward
    datatype transport = Only of ptype
                       | Any of ptype (* ptype gives your current mode *)

    type step = (nid * ptype)
    type path = step list

    type graph

    (* converts a list of nodes and a list of edges into a graph *)
    val tograph : 
        (* nodes *)
        (string * Types.node_tag * int * int) list *
        (* edges *)
        (string * string * Types.edge_tag) list
        -> graph
        
    (* get the original edge or node list that made a graph --
       note that the location fields are always (0, 0) in the 
       returned list, because we don't save it. *)
    val orig_edges : graph -> Types.edge list
    val orig_nodes : graph -> Types.node list

    val name_to_nid : graph -> string -> nid
    val nid_to_name : graph -> nid -> string

    (* returns the set of adjacent nodes along with the type of edge you need
       to follow to reach each one *)
    val adjacent : graph -> direction -> transport -> nid -> step list

    (* returns all nodes reachable from a set of nodes through valid moves *)
    val reachable : graph -> direction -> transport -> nid list -> int
                    -> nid list

    (* distance g tr src dst
       
       assumes all nodes are reachable from one another.
       (otherwise, result will be a very high distance)
       return the distance from src to dst (when driving a
       'transport'). *)
    val distance : graph -> transport -> nid -> nid -> int

    val shortest_path : graph -> transport -> nid -> nid -> path option

    val shortest_path_distance :
        graph -> transport -> nid -> nid -> (int * path) option

    (* number of nodes n; all nids in this graph will then be in the
       range 0..n-1 with no gaps *)
    val length : graph -> int

    (* bank, hq, robberstart, or ordinary? *)
    val nodetype : graph -> nid -> Types.node_tag

    val banks : graph -> nid list
    val hq : graph -> nid
    val robberstart : graph -> nid


    (* sometimes you need a dummy graph for initialization,
       but don't call any functions on it! *)
    val dummy : graph
end
