package edu.cmu.cs.icfp2005.dfa;

import java.io.*;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class DeadlyFutureArena {

    final private static String DIR_PREFIX = "/afs/andrew.cmu.edu/usr1/agoode/icfp2005/dfa2000/";

    // final private static String DIR_PREFIX = "./dfa2000/";

    // final public static String COPS_DIR = DIR_PREFIX + "cops";

    // final public static String ROBBERS_DIR = DIR_PREFIX + "robbers";

    final public static String SOUND_FILE = DIR_PREFIX + "coin.wav";

    final public static String RESULTS_DIR = DIR_PREFIX + "results";

    final public static String WEB_DIR = "/afs/cs.cmu.edu/user/agoode/www/dfa2000";

    // final public static String WEB_DIR = ".";

    // final public static String OUTPUT = DIR_PREFIX + "bots-list.sml";

    // final public static String CM_OUTPUT = DIR_PREFIX + "bots.cm";

    final public static String ERROR_LOG_NAME = RESULTS_DIR
            + "/build-errors.txt";

    // final private static Pattern pat = Pattern
    // .compile("structure\\s+(\\S+)\\s+:>\\s+(COP|ROBBER)");
    //
    // final private static MessageDigest digester;
    // static {
    // MessageDigest m = null;
    // try {
    // m = MessageDigest.getInstance("SHA1");
    // } catch (NoSuchAlgorithmException e) {
    // e.printStackTrace();
    // System.exit(0);
    // }
    // digester = m;
    // }

    public static void main(String[] args) {
        // get directory to watch
        // File copsDir = new File(COPS_DIR);
        // File robbersDir = new File(ROBBERS_DIR);

        DeadlyFutureArena dfa2001 = new DeadlyFutureArena();
        System.out.println("Entering Deadly Future Arena 2001...");
        dfa2001.run();
    }

    private long lastErase;

    private void run() {
        while (true) {
            try {
                // System.out.println("updating...");
                checkResultsAndWriteSummary();
            } catch (IOException e) {
                e.printStackTrace();
            }

            System.out.print("\r");
            try {
                Thread.sleep(200);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            System.out.print("Up to date, waiting for new results");

            try {
                Thread.sleep(2000);
            } catch (InterruptedException e1) {
                e1.printStackTrace();
            }
        }
    }

    private void eraseSomeOld(File results[]) {
        if (results.length == 0) {
            return;
        }

        File oldest = results[0];
        long oldestTime = oldest.lastModified();

        for (int i = 1; i < results.length; i++) {
            File file = results[i];
            long t = file.lastModified();
            if (t < oldestTime) {
                oldestTime = t;
                oldest = file;
            }
        }

        long minutes = (System.currentTimeMillis() - oldestTime) / 60000;
        if (oldest.length() > 0 && minutes > 40) {
            File oldestErr = new File(oldest.getPath() + ".err");

            System.out.println();
            System.out.println("Erasing " + oldest + " (" + minutes
                    + " minutes old)");

            oldest.delete();
            oldestErr.delete();
        }
    }

    static final private Pattern scorePat = Pattern
            .compile("^scores\\s+\\((.+)\\)$");

    static final private Pattern sexpPat = Pattern
            .compile("\\(\"([^\"]+)\"\\s+\"([^\"]+)\"\\s+(\\d+)\\)");

    static final private Pattern filenamePat = Pattern
            .compile("^([^-]+)-([^-]+)-([^-]+)-([^-]+)-([^-]+)-([^-]+)\\.out$");

    // TIMER: move-6rob= cnt: 0 avg: 0.000000 tlt: 0.000000 min:
    // 1119844834.120636 max: 0.000000
    static final private Pattern timerPat = Pattern
            .compile("^TIMER: move-(\\d)(cop|rob)=\\s+cnt:\\s+([0-9]+)"
                    + ".*tlt:\\s+([0-9.]+)");

    private void checkResultsAndWriteSummary() throws IOException {
        Map robberScores = new HashMap();
        Map copScores = new HashMap();

        // System.out.println("reading results");

        File resultsDir = new File(RESULTS_DIR);
        File results[] = resultsDir.listFiles(outFilter);

        SortedMap errors = new TreeMap(Collections.reverseOrder());

        for (int i = 0; i < results.length; i++) {
            File result = results[i];

            String botNames[] = new String[6];
            int botMoveCounts[] = new int[6];
            double botTotalTime[] = new double[6];
            Matcher botNameMatcher = filenamePat.matcher(result.getName());
            botNameMatcher.matches();
            for (int j = 0; j < botNames.length; j++) {
                botNames[j] = botNameMatcher.group(j + 1);
            }

            // read err
            StringBuffer err = new StringBuffer();
            File errFile = new File(result.getPath() + ".err");
            try {
                FileReader fr = new FileReader(errFile);
                BufferedReader errReader = new BufferedReader(fr);
                String errLine;
                while ((errLine = errReader.readLine()) != null) {
                    if (errLine.startsWith("TIMER:")) {
                        // System.out.println(errLine);
                        // System.out.println(timerPat);
                        Matcher m = timerPat.matcher(errLine);
                        m.find();
                        int bot = Integer.parseInt(m.group(1)) - 1;
                        botMoveCounts[bot] = Integer.parseInt(m.group(3));
                        botTotalTime[bot] = Double.parseDouble(m.group(4));
                    } else if (true || (!errLine.startsWith("DEBUG:") && !errLine
                            .startsWith("ECHO:"))) {
                        err.append(errLine);
                        err.append("\n");
                    }
                }

                if (err.length() > 0) {
                    long lm = errFile.lastModified();
                    String dStr = dFormat.format(new Date(lm));
                    String errStr = dStr + "\n\n";
                    errors.put(new Long(lm), new String[] { result.getName(),
                            errStr + err.toString() });
                }

                errReader.close();
                fr.close();
            } catch (FileNotFoundException e) {
            }

            // read output
            FileReader fr = new FileReader(result);
            BufferedReader b = new BufferedReader(fr);

            String line = b.readLine();
            if (line == null) {
                continue;
            }
            // scores (("1cop" "copA" 72) ("2cop" "copB" 12) ("3cop" "copC" 12)
            // ("4cop" "copD" 12) ("5cop" "copE" 12) ("6rob" "robber" 0))
            Matcher m = scorePat.matcher(line.trim());
            if (m.matches()) {
                String item = m.group(1);
                Matcher m2 = sexpPat.matcher(item.trim());
                for (int j = 1; j <= 6; j++) {
                    m2.find();
                    String botName = botNames[j - 1];
                    // String botName = m2.group(1);
                    int score = Integer.parseInt(m2.group(3));
                    // System.out.println(bot + " -> " + score);

                    Map scores;
                    if (j == 6) {
                        scores = robberScores;
                    } else {
                        scores = copScores;
                    }
                    Score s = (Score) scores.get(botName);
                    if (s == null) {
                        s = new Score(botName);
                    }
                    s.addGame(score, botMoveCounts[j - 1], botTotalTime[j - 1]);
                    scores.put(botName, s);
                }
            }
            b.close();
            fr.close();
        }
        writeSummary(robberScores, copScores, errors);

        if (System.currentTimeMillis() - lastErase > 1000 * 60 * 3) {
            eraseSomeOld(results);
            lastErase = System.currentTimeMillis();
        }
    }

    final static public String rfc2822FormatString = "EEE, d MMM yyyy HH:mm:ss Z";

    final private DateFormat dFormat = new SimpleDateFormat(rfc2822FormatString);

    final private NumberFormat ppgFormat = new DecimalFormat("#.##");

    final private NumberFormat statFormat = new DecimalFormat("#.##");

    final private NumberFormat timerFormat = new DecimalFormat("#.######");

    private String colors[] = { "red", "green", "blue" };

    private boolean justBody = true;

    private void writeSummary(Map robberScores, Map copScores, SortedMap errors)
            throws IOException {
        // System.out.println("writing summary");

        Collections.shuffle(Arrays.asList(colors));

        File out = File.createTempFile("deadlyfuture", ".html", new File(
                WEB_DIR));
        out.deleteOnExit();

        FileWriter fw = new FileWriter(out);
        PrintWriter p = new PrintWriter(fw);

        if (!justBody) {
            p.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                    + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\"\n"
                    + "    \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">");
            p
                    .println("<html xmlns=\"http://www.w3.org/1999/xhtml\""
                            + " xml:lang=\"en\" >\n"
                            + "<head><meta http-equiv=\"REFRESH\" content=\"10\" />"
                            + "<style type=\"text/css\">\n"
                            + " H1 { font: bold 24pt Verdana,Arial,Helvetica,sans-serif }\n"
                            + " P  { font: 12pt Verdana,Arial,Helvetica,sans-serif }\n"
                            + " TH { font: bold 12pt Verdana,Arial,Helvetica,sans-serif ; \n"
                            + "      text-align: center}\n"
                            + " TD { font: 12pt Verdana,Arial,Helvetica,sans-serif }\n"
                            + " B  { font: bold 14pt Verdana,Arial,Helvetica,sans-serif }\n"
                            + " .red { color : #880000 }\n"
                            + " .green { color : #008800 }\n"
                            + " .blue { color : #000088 }\n"
                            + " .total { font-size : x-large }\n"
                            + "</style>\n"
                            + "<title>DEADLY FUTURE ARENA 2001</title>"
                            + "</head><body>");
        }

        p.println("<h1>Welcome to <span class=\"" + colors[0]
                + "\">DEADLY</span>" + " <span class=\"" + colors[1]
                + "\">FUTURE</span>" + " <span class=\"" + colors[2]
                + "\">ARENA</span> 2001</h1>");
        p.print("<h2>Scores at ");
        p.print(dFormat.format(new Date(System.currentTimeMillis())));
        p.println("</h2>");

        printTable(p, "Robbers", robberScores);
        printTable(p, "Cops", copScores);

        printBuildErrors(p);

        printRunErrors(errors, p);

        if (!justBody) {
            p.println("</body></html>");
        }

        p.close();
        fw.close();

        String filename = WEB_DIR + (justBody ? "/index.body" : "/index.html");
        File indexFile = new File(filename);
        out.renameTo(indexFile);
    }

    private void printRunErrors(SortedMap errors, PrintWriter p) {
        if (!errors.isEmpty()) {
            p.println("<h2>Run Errors</h2>");
        }

        for (Iterator iter = errors.keySet().iterator(); iter.hasNext();) {
            Object key = iter.next();
            String ss[] = (String[]) errors.get(key);
            String h = ss[0];
            String body = ss[1];

            p.println("<h3>" + h.substring(0, h.length() - 4) + "</h3>");
            p.print("<pre>");
            p.println(body);
            p.print("</pre>");
        }
    }

    private void printBuildErrors(PrintWriter p) throws FileNotFoundException,
            IOException {
        p.println("<h2>Build Errors</h2>");
        p.print("<pre>");

        FileReader fr = new FileReader(ERROR_LOG_NAME);
        BufferedReader errLog = new BufferedReader(fr);
        String line;
        boolean noErr = true;
        while ((line = errLog.readLine()) != null) {
            noErr = false;
            p.println(line);
        }
        errLog.close();
        fr.close();

        if (noErr) {
            p.println("No errors!");
        }

        p.println("</pre>");
    }

    private void printTable(PrintWriter p, String type, Map scores) {
        p.println("<h2>" + type + "</h2>");
        Collection c = scores.values();
        Score s[] = new Score[c.size()];
        c.toArray(s);
        List l = Arrays.asList(s);
        Collections.sort(l);
        Collections.reverse(l);

        p.println("<table border=\"2\" cellpadding=\"4\" cellspacing=\"0\" "
                + "style=\"margin: 1em 1em 1em 0; background: #f9f9f9; "
                + "border: 1px #aaa solid; border-collapse: collapse\">\n"
                + "<colgroup><col /><col /><col align=\"char\" char=\".\" />"
                + "<col /><col align=\"char\" char=\".\" />" + "</colgroup>\n"
                + "<tr><th>Name</th><th>Games Played</th>"
                + "<th>Seconds Per Move</th>" + "<th>Total Points</th>"
                + "<th>Points Per Game</th></tr>");

        for (int i = 0; i < s.length; i++) {
            Score ss = s[i];
            p.print("<tr>");
            p.print("<td>" + ss.getName() + "</td>");
            p.print("<td>" + ss.getNumGames() + "</td>");
            p.print("<td>" + timerFormat.format(ss.getTimePerMove()) + "</td>");
            p.print("<td>" + ss.getTotalScore() + "</td>");
            p.print("<td>" + ppgFormat.format(ss.getScorePerGame()) + " ± "
                    + statFormat.format(ss.getScoreSE()) + "</td>");
            p.println("</tr>");
        }

        p.println("</table>");
    }

    public static class MyFileFilter implements FileFilter {
        final private String ext;

        public boolean accept(File pathname) {
            String lc = pathname.getName().toLowerCase();
            return pathname.isFile() && (lc.endsWith(ext));
        }

        public MyFileFilter(String ext) {
            this.ext = ext.toLowerCase();
        }
    }

    public int fact(int n) {
        if (n == 1 || n == 0) {
            return 1;
        } else if (n < 0) {
            return 0;
        } else {
            return n * fact(n - 1);
        }
    }

    final public static FileFilter smlFilter = new MyFileFilter(".sml");

    final public static FileFilter outFilter = new MyFileFilter(".out");
}