signature MATRIX =
sig

  type connects = real Vector.vector Vector.vector
  type beliefs = real Array.array

  (* Print out a connection network using Spew. *)
  val print_connects : connects -> unit
  (* Print out a series of beliefs values using Spew. *)
  val print_beliefs : beliefs -> unit

  (* Don't try and do anything useful with these values. *)
  val dummy_connects : connects
  val dummy_beliefs : beliefs

  (* Build an adjacency network from the given graph. *)
  val create_adjacent_connects : Graph.graph -> Graph.transport -> connects

  (* Build a reachability-in-one-step network from the given graph. *)
  val create_reach_connects : Graph.graph -> Graph.transport -> connects

  (* Build a weighted adjacency matrix where edges are weighted by the degree
   * of their source node. *)
  val create_spread_connects : Graph.graph -> Graph.transport -> connects

  (* Build a belief set of the specified length. *)
  val new_beliefs : int -> beliefs
  (* Zero out a set of beliefs. *)
  val clear_beliefs : beliefs -> unit

  (* Given a set of connections C, beliefs D, and beliefs B, compute
      D := C*B
     Note that D and B should be distinct sets of beliefs. *)
  val multiply_with_dest : connects -> beliefs -> beliefs -> unit

  (* Given beliefs D and B, compute the pointwise sum,
      D := D+B
     Storing the result in D.
   *)
  val add_in_place : beliefs -> beliefs -> unit

  (* Given beliefs D and B and scalar s, compute the pointwise sum,
      D := D+(B*s)
     Storing the result in D.
   *)
  val add_in_place_with_scale : beliefs -> beliefs -> real -> unit
end
