structure Matrix :> MATRIX =
struct

  structure G = Graph
  structure V = Vector
  structure A = Array

  val toString = Real.fmt (StringCvt.FIX (SOME 2))

  type connects = real Vector.vector Vector.vector
  type beliefs = real Array.array

  fun print_connects m =
      Util.for 0 (V.length m - 1)
      (fn i => let val r = V.sub (m, i)
                   val _ = Util.for 0 (V.length r - 1)
                            (fn j => Spew.spe ((toString (V.sub (r, j))) 
                                               ^ " "))
                in
                  Spew.spe "\n"
                end)

  fun print_beliefs v = 
      (Util.for 0 (A.length v - 1)
       (fn i => Spew.spe ((toString (A.sub (v, i))) ^ " "));
       Spew.spe "\n")

  val dummy_connects = V.fromList [V.fromList [0.0]]
  val dummy_beliefs = A.fromList [0.0]

  fun create_adjacent_connects g trans =
      let val length = G.length g
          fun create_row nd = 
              let val adjacent = G.adjacent g G.Forward trans nd
                  fun create_cell nd' = if List.exists
                                               (fn (nd'', _) => nd' = nd'')
                                               adjacent
                                        then 1.0 else 0.0

              in
                V.tabulate (length, create_cell)
              end
      in
        V.tabulate (length, create_row)
      end

  fun create_reach_connects g trans =
      let val length = G.length g
          fun create_row nd = 
              let val reachable = G.reachable g G.Forward trans [nd] 1
                  fun create_cell nd' = if List.exists (fn nd'' => nd' = nd'') 
                                                       reachable
                                        then 1.0 else 0.0

              in
                V.tabulate (length, create_cell)
              end
      in
        V.tabulate (length, create_row)
      end

  fun create_spread_connects g trans =
      let val length = G.length g
          val rlist = V.tabulate (length, fn nd => G.reachable g G.Forward
                                                               trans [nd] 1)
          fun create_row nd =
              let val reachable = V.sub (rlist, nd)
                  fun create_cell nd' = if List.exists (fn nd'' => nd' = nd'')
                                                       reachable
                                        then let val r = V.sub (rlist, nd')
                                                 val n = List.length r
                                             in
                                               1.0 / (Real.fromInt n)
                                             end
                                        else 0.0
              in
                V.tabulate (length, create_cell)
              end
      in
        V.tabulate (length, create_row)
      end

  fun new_beliefs length = A.tabulate (length, fn _ => 0.0)

  fun clear_beliefs a = A.modify (fn _ => 0.0) a

  fun multiply_with_dest m r a = 
      Util.for 0 (A.length a - 1)
      (fn i => A.update (r, i, A.foldli (fn (j, w, r) => r +
                                              (V.sub (V.sub (m, i), j)) * w)
                                        0.0 a))

  fun add_in_place d a =
      A.modifyi (fn (i, v:real) => v + A.sub (a, i)) d

  fun add_in_place_with_scale d a s =
      A.modifyi (fn (i, v:real) => v + (s * A.sub (a, i))) d
end
