structure Message =
struct

type pid = Types.pid
type bid = Types.bid
type nid = string
type time = Types.time

datatype ptype = datatype Types.ptype
datatype node_tag = datatype Types.node_tag
datatype edge_tag = datatype Types.edge_tag

type node = nid * node_tag * int * int
type edge = nid * nid * edge_tag

type register_msg = pid * ptype

type world_skeleton_msg  = {name : pid, 
                            robber: pid,
                            cop : pid * pid * pid * pid * pid, 
                            graph : Graph.graph }

type game_over_msg = unit

type world_msg = {time: time,
                  loot: int,
                  dirty_cops : bid list,
                  controlling : (bid * pid) list,
                  accused : (bid * bid * time) list,
                  bank_vals: (nid * int) list,
                  evidence: (nid * time) list,
                  smell: int option,
                  player_locs : (pid * nid * ptype) list}

type move_msg = nid * ptype * pid

datatype offer = Turncoat | StraightArrow

type clean_cop_move_msg = {moves: move_msg list,
                           offer : offer,
                           accuse : pid list}

type dirty_cop_move_msg = move_msg list

type accused = unit

datatype bribe_msg = Bribe | NoBribe

datatype offered_cops_msg = NotTelling
                          | Offered of bid list

datatype bribe_result = Push of pid * nid
                      | BribeChoose of bid list
                      | BribeNone

type robber_move_msg = {move : move_msg,
                        result : bribe_result}

type inform_msg = (pid * nid * ptype * time * int) list

type plan_msg = (pid * nid * ptype * time) list

type vote_msg = bid list

type vote_res_msg = bid option

type from_inform_msg = (pid * inform_msg) list
type from_plan_msg = (bid * plan_msg) list
        
type state = exn

type dirtycop = {
  inform : world_msg * state -> inform_msg * state,
  plan : from_inform_msg * state -> plan_msg * state,
  vote : from_plan_msg * state -> vote_msg * state,

  move : vote_res_msg * state -> dirty_cop_move_msg * state,

  (* For forming dastardly plans! *)
  r_inform : world_msg * state -> inform_msg * state,
  r_plan : from_inform_msg * state -> plan_msg * state,
  r_vote : from_plan_msg * state -> vote_msg * state,
  r_results : vote_res_msg * state -> state
}

type cleancop = {
  register : unit -> register_msg * state,
  world_skeleton : world_skeleton_msg * state -> state,


  (* note: state usually has a different tag, 
     since it is now a state for a dirty cop. *)
  dirtify : state -> dirtycop * state,

  inform : world_msg * state -> inform_msg * state,
  plan : from_inform_msg * state -> plan_msg * state,
  vote : from_plan_msg * state -> vote_msg * state,

  move : vote_res_msg * state -> clean_cop_move_msg * state
}

type robber = {
  register : unit -> register_msg * state,
  world_skeleton : world_skeleton_msg * state -> state,

  inform : world_msg * state -> inform_msg * state,
  plan : from_inform_msg * state -> plan_msg * state,
  vote : from_plan_msg * state -> vote_msg * state,
  bribe : vote_res_msg * state -> bribe_msg * state,
  move : offered_cops_msg * state -> robber_move_msg * state
}

end
