structure Rand :> RAND = struct

(* initialize the random number generator *)
fun init () =
    let open Word32
        infix xorb (* note use of infix for added clarity *)
        (* SHA1 hash of our original submission -- a secure seed *)
        val seed = 0wx8cb71aba
              xorb 0wx2c57dcbf
              xorb 0wx43b4d5e3
              xorb 0wx6668b65e
              xorb 0wx0440645c
              (* xor, for added security *)
    in
        MLton.Random.srand seed;
        seed
    end

(* generate a random int in [0,n-1] *)
fun randN n =
    let val r = MLton.Random.rand()
        val max = Word.- (0w0, 0w1)
        val toReal = Real.fromLargeInt o Word.toLargeInt
    in Real.floor (((toReal r) * (Real.fromInt n)) / (toReal max))
    end

(* simpler *)
fun tomrand n =
    let 
        val r = MLton.Random.rand ();
        val r = Word.>> (r, 0w5)
        val r = Word.andb(r, 0wxFFFFFF)
    in
        Word.toInt r mod n
    end

fun permute_list lst =
    let val a = Array.fromList lst
        val len = Array.length a
        fun to_list a = Array.foldr (op::) [] a
        fun swap (i, j, a) =
            let val a_i = Array.sub (a, i)
                val a_j = Array.sub (a, j)
            in
                Array.update (a, i, a_j);
                Array.update (a, j, a_i)
            end
        fun loop i =
            if i < len then
                (swap (i, randN len, a);
                 loop (i+1))
            else
                to_list a
    in
        loop 0
    end
end
