structure ServerCommon =
struct

open Parsing

fun int_to_string i = if i > 0 then Int.toString i 
                      else "-" ^ (Int.toString (~i))

(* NB: doing send_* here instead of send o unparse_*, since these are lists
   and unparsing a big string might be inefficient -- and all we do is send
   it anyway *)
fun send_list name unparse_elt elts =
    let val start = name ^ "\\"
        val finish = name ^ "/"
        val delim = name ^ ": "
    in
        send start;
        List.app (fn elt => send (delim ^ unparse_elt elt)) elts;
        send finish
    end

fun send_inform_msg inf_msg =
    send_list "inf"
              (fn (pid, nid, ptype, time, cert) =>
                     (pid ^ " "
                    ^ nid ^ " "
                    ^ unparse_ptype ptype ^ " "
                    ^ int_to_string time ^ " "
                    ^ int_to_string cert)) 
              inf_msg

fun send_plan_msg plan_msg =
    send_list "plan"
              (fn (pid, nid, ptype, time) =>
                    (pid ^ " " ^ nid ^ " " ^ unparse_ptype ptype ^ " "
                    ^ int_to_string time))
              plan_msg

fun send_vote_msg vote_msg =
    send_list "vote" (fn pid => pid) vote_msg

fun recv_inform_msg () =
    recv_list "inf" (fn [bot, loc, ptype, world, cert] =>
                            (bot, loc, parse_ptype ptype,
                             valOf (Int.fromString world),
                             valOf (Int.fromString cert))
                       | _ => raise Fail "recv_inform_msg: wrong # of args")

fun recv_plan_msg () =
    recv_list "plan" (fn [bot, loc, ptype, world] =>
                            (bot, loc, parse_ptype ptype,
                             valOf (Int.fromString world))
                       | _ => raise Fail "recv_plan_msg: wrong # of args")

fun recv_vote_msg () =
    recv_list "vote" (fn [bot] => bot
                       | _ => raise Fail "recv_vote_msg: wrong # of args")

fun recv_from_inform_msg () =
    recv_list "from" (fn [bot] => (bot, recv_inform_msg ())
                      | _ => raise Fail "recv_from_inform_msg: expected bot")

fun recv_from_plan_msg () =
    recv_list "from" (fn [bot] => (bot, recv_plan_msg ())
                      | _ => raise Fail "recv_from_plan_msg: expected bot")

fun recv_vote_res_msg () =
    case tokens (recv ()) of
        ["winner:", bot] => SOME bot
      | ["nowinner:"] => NONE
      | _ => raise Fail "recv_vote_res_msg: expected \"winner:\"/\"nowinner:\""

end
