(* This turns a value of type 'cop' into an actual process that does
   I/O to communicate with the server. *)

structure ServerCoploop :> COPLOOP =
struct

open Parsing
open ServerCommon

val register_timing = Spew.new_timing ()
val world_skel_timing = Spew.new_timing ()
val inform_timing = Spew.new_timing ()
val plan_timing = Spew.new_timing ()
val vote_timing = Spew.new_timing ()
val move_timing = Spew.new_timing ()
val r_inform_timing = Spew.new_timing ()
val r_plan_timing = Spew.new_timing ()
val r_vote_timing = Spew.new_timing ()
val r_results_timing = Spew.new_timing ()

fun run (cop : Message.cleancop) =
    let val botname = !Super.BOTNAME

        val (r_msg, s) = Spew.time register_timing (#register cop) ()
        val _ = Spew.report_timing [("register-" ^ botname, register_timing)]
        val () = send (unparse_register_msg r_msg)

        val ws_msg = recv_world_skeleton ()
        val s =  Spew.time world_skel_timing (#world_skeleton cop) (ws_msg, s)
        val _ = Spew.report_timing [("world_skel-" ^ botname, 
                                     world_skel_timing)]
        val name = #name ws_msg

        fun dirty_loop (cop : Message.dirtycop, s) w_msg_o=
            (case w_msg_o of
               NONE => ()
             | SOME w_msg =>
               let val (inf_msg, s) = Spew.time inform_timing (#inform cop) 
                                                (w_msg, s)
 		   val _ = Spew.report_timing [("inform-" ^ botname, 
						inform_timing)]
                   val () = send_inform_msg inf_msg
                            
                   val frominf_msg = recv_from_inform_msg ()
                   val (plan_msg, s) = Spew.time plan_timing (#plan cop)
                                                 (frominf_msg, s)
 		   val _ = Spew.report_timing [("plan-" ^ botname, 
						plan_timing)]
                   val () = send_plan_msg plan_msg
                            
                   val fromplan_msg = recv_from_plan_msg ()
                   val (vote_msg, s) = Spew.time vote_timing
                                            (#vote cop) (fromplan_msg, s)
 		   val _ = Spew.report_timing [("vote-" ^ botname, 
						vote_timing)]
                   val () = send_vote_msg vote_msg
                            
                   val voteres_msg = recv_vote_res_msg ()
                   val (mov_msg, s) = Spew.time move_timing
                                        (#move cop) (voteres_msg, s)
 		   val _ = Spew.report_timing [("move-" ^ botname, 
						move_timing)]

                   val () = send (unparse_dirty_cop_move_msg mov_msg)
                            
                   val w_msg_o = recv_world_or_game_over ()
                in
                    dirty_loop2 (cop, s) w_msg_o
                end)

        and dirty_loop2 (cop, s) w_msg_o =
           (case w_msg_o of
               NONE => ()
             | SOME w_msg =>
               let val (inf_msg, s) = Spew.time r_inform_timing
                                                (#r_inform cop) (w_msg, s)
 		   val _ = Spew.report_timing [("r_inform-" ^ botname, 
						r_inform_timing)]
                   val () = send_inform_msg inf_msg
                            
                   val frominf_msg = recv_from_inform_msg ()
                   val (plan_msg, s) = Spew.time r_plan_timing (#r_plan cop)
                                                 (frominf_msg, s)
 		   val _ = Spew.report_timing [("r_plan-" ^ botname, 
						r_plan_timing)]
                   val () = send_plan_msg plan_msg
                             
                   val fromplan_msg = recv_from_plan_msg ()
                   val (vote_msg, s) = Spew.time r_vote_timing (#r_vote cop)
                                                 (fromplan_msg, s)
 		   val _ = Spew.report_timing [("r_vote-" ^ botname, 
						r_vote_timing)]
                   val () = send_vote_msg vote_msg

                   val voteres_msg = recv_vote_res_msg ()
                   val s = Spew.time r_results_timing (#r_results cop)
                                                      (voteres_msg, s)
 		   val _ = Spew.report_timing [("r_results-" ^ botname, 
						r_results_timing)]

                   val w_msg_o = recv_world_or_game_over ()
                in 
                  dirty_loop (cop, s) w_msg_o
                end)

        fun clean_loop (cop : Message.cleancop) s =
            let val w_msg_o = recv_world_or_game_over ()
            in
              case w_msg_o of
                NONE => ()
              | SOME w_msg =>
                if List.exists (fn bd => bd = name) (#dirty_cops w_msg)
                then dirty_loop ((#dirtify cop) s) w_msg_o
                else 
                  let  val (inf_msg, s) = Spew.time inform_timing
                                                    (#inform cop) 
                                                    (w_msg, s)
 		       val _ = Spew.report_timing [("inform-" ^ botname, 
						    inform_timing)]
                       val () = send_inform_msg inf_msg
                                
                       val frominf_msg = recv_from_inform_msg ()
                       val (plan_msg, s) = Spew.time plan_timing (#plan cop)
                                                     (frominf_msg, s)
 		       val _ = Spew.report_timing [("plan-" ^ botname, 
						    plan_timing)]
                       val () = send_plan_msg plan_msg
                                
                       val fromplan_msg = recv_from_plan_msg ()
                       val (vote_msg, s) = Spew.time vote_timing (#vote cop)
                                                     (fromplan_msg, s)
 		       val _ = Spew.report_timing [("vote-" ^ botname, 
						    vote_timing)]
                       val () = send_vote_msg vote_msg
                                
                       val voteres_msg = recv_vote_res_msg ()
                       val (mov_msg, s) = Spew.time move_timing (#move cop)
                                                    (voteres_msg, s)
 		       val _ = Spew.report_timing [("move-" ^ botname, 
						    move_timing)]
                       val () = send (unparse_clean_cop_move_msg mov_msg)
                  in
                    clean_loop cop s
                  end
            end
    in
      clean_loop cop s
    end

end
