(* This turns a value of type 'robber' into an actual process that
   does I/O to communicate with the server. *)

structure ServerRobberloop :> ROBBERLOOP =
struct

open Parsing
open ServerCommon

val register_timing = Spew.new_timing ()
val world_skel_timing = Spew.new_timing ()
val inform_timing = Spew.new_timing ()
val plan_timing = Spew.new_timing ()
val bribe_timing = Spew.new_timing ()
val vote_timing = Spew.new_timing ()
val move_timing = Spew.new_timing ()

fun run (robber : Message.robber) =
    let val botname = !Super.BOTNAME

        val (r_msg, s) = Spew.time register_timing (#register robber) ()
        val _ = Spew.report_timing [("register-" ^ botname, register_timing)]
        val () = send (unparse_register_msg r_msg)

        val ws_msg : Message.world_skeleton_msg = recv_world_skeleton ()
        val s = Spew.time world_skel_timing (#world_skeleton robber)
                          (ws_msg, s)
        val _ = Spew.report_timing [("world_skel-" ^ botname, 
                                     world_skel_timing)]
        fun loop s =
            let val w_msg_o = recv_world_or_game_over ()
            in
                case w_msg_o of
                    NONE => ()
                  | SOME w_msg =>
                        let val (inf_msg, s) = Spew.time inform_timing
                                                         (#inform robber) 
                                                         (w_msg, s)
 			    val _ = Spew.report_timing [("inform-" ^ botname, 
							 inform_timing)]
                            val () = send_inform_msg inf_msg

                            val frominf_msg = recv_from_inform_msg ()
                            val (plan_msg, s) = Spew.time plan_timing
                                                          (#plan robber)
                                                          (frominf_msg, s)
 			    val _ = Spew.report_timing [("plan-" ^ botname, 
							 plan_timing)]
                            val () = send_plan_msg plan_msg

                            val fromplan_msg = recv_from_plan_msg ()
                            val (vote_msg, s) = Spew.time vote_timing 
                                                          (#vote robber)
                                                          (fromplan_msg, s)
 			    val _ = Spew.report_timing [("vote-" ^ botname, 
							 vote_timing)]
                            val () = send_vote_msg vote_msg

                            val voteres_msg = recv_vote_res_msg ()
                            val (b_msg, s) = Spew.time bribe_timing 
                                                       (#bribe robber)
                                                       (voteres_msg, s)
 			    val _ = Spew.report_timing [("bribe-" ^ botname, 
							 bribe_timing)]
                            val () = send (unparse_bribe_msg b_msg)

                            val oc_msg = recv_offered_cops ()
                            val (m_msg, s) = Spew.time move_timing 
                                                       (#move robber)
                                                       (oc_msg, s)
			    val _ = Spew.report_timing [("move-" ^ botname, 
							 move_timing)]
                        in
                            send (unparse_robber_move_msg m_msg);
                            loop s
                        end
            end
    in
        loop s
    end

end
