structure Spew =
  struct
    val ECHOING = Params.flag false (SOME ("-echoing", 
                                           "Enable echoing output"))
                             "echoing"
    val ECHO_FILE = Params.param "echo" (SOME ("-echo_file", 
                                               "Enable output file"))
                             "echo-file"
    val TIMING = Params.flag false (SOME ("-timing", "Enable timing output"))
                             "timing"
    val SPEWING = Params.flag false (SOME ("-spewing", 
                                           "Enable generic debug output"))
                             "spewing"

    (* No DEBUG prefix or newline. *)
    fun spe s =
        if !SPEWING then
            (
            TextIO.output (TextIO.stdErr, s);
            TextIO.flushOut TextIO.stdErr
            )
        else
            ()

    fun spew s =
        if !SPEWING then
            (
            TextIO.output (TextIO.stdErr, "DEBUG: ");
            TextIO.output (TextIO.stdErr, s);
            TextIO.output (TextIO.stdErr, "\n");
            TextIO.flushOut TextIO.stdErr
            )
        else
            ()

    type timing = {cnt : int,
                   tlt : Time.time,
                   min : Time.time,
                   max : Time.time} ref
               
    fun new_timing () : timing = ref {cnt=0, tlt=Time.zeroTime, 
                                      min=Time.now (), max=Time.zeroTime}

    fun time (t:timing) f x =
        let val t1 = Time.now () (* Timer.startRealTimer () *)
            val result = f x
            val t2 = Time.now () (* Timer.checkRealTimer t1 *)
            val tim = !t
	    val cnt = (#cnt tim) + 1
            val cur = Time.- (t2, t1)
            val tlt = Time.+ (#tlt tim, cur)
            val min = if Time.< (cur, #min tim) then cur else #min tim
            val max = if Time.> (cur, #max tim) then cur else #max tim
            val _ = t := {cnt=cnt, tlt=tlt, min=min, max=max}
        in
          result
        end

    fun report_timing ((n, (t : timing))::ts) = 
	if !TIMING then
            let val total : LargeInt.int = Time.toMicroseconds (#tlt (!t))
		val count : LargeInt.int = LargeInt.fromInt (#cnt (!t))
		val avg = LargeInt.div (total, if count = 0 then 1 else count)
		val avg = Time.fromMicroseconds avg
		val tlt = (#tlt (!t))
		val fmt = Time.fmt 6
		val _ = TextIO.output 
			    (TextIO.stdErr, "TIMER: " ^ n 
					    ^ "= cnt: " 
					    ^ (LargeInt.toString count)
					    ^ " avg: " ^ (fmt avg)
					    ^ " tlt: " ^ (fmt tlt)
					    ^ " min: " ^ (fmt (#min (!t)))
					    ^ " max: " ^ (fmt (#max (!t)))
					    ^ "\n")
		val _ = TextIO.flushOut TextIO.stdErr
            in
		report_timing ts
            end
	else ()
      | report_timing nil = ()

    val echoOut = ref NONE : TextIO.outstream option ref

    fun echo s =
        if !ECHOING then
          let val out = case !echoOut 
                         of SOME out => out
                          | NONE => let val file = !ECHO_FILE
                                        val out = 
                                            if file = "" 
                                            then TextIO.stdErr
                                            else TextIO.openOut file
                                        val _ = echoOut := SOME out
                                    in
                                      out
                                    end
          in
            TextIO.output (out, "ECHO: ");
            TextIO.output (out, s);
            TextIO.flushOut out
          end
        else
          ()

  end


