signature SUPER =
sig

  type super

  type pid = Types.pid
  type bid = Types.bid
  type node = Types.node
  type edge = Types.edge
  type time = Types.time
  type nid = Types.nid
  type ptype = Types.ptype

  type register_msg = pid * ptype

  type world_skeleton_msg  = {name : pid, 
			      robber: pid,
			      cop : pid * pid * pid * pid * pid, 
                              graph : Graph.graph }

  type world_msg = {time: time,
		    loot: int,
                    dirty_cops : bid list,
                    controlling : (bid * pid) list,
                    accused : (bid * bid * time) list,
		    bank_vals: (nid * int) list,
		    evidence: (nid * time) list,
		    smell: int option,
		    player_locs : (pid * nid * ptype) list}

  type move_msg = nid * ptype * pid

  datatype offer = datatype Message.offer

  type clean_cop_move_msg = {moves: move_msg list,
                             offer : offer,
                             accuse : pid list}

  type dirty_cop_move_msg = move_msg list

  type accused = unit

  datatype bribe_msg = datatype Message.bribe_msg

  datatype offered_cops_msg = datatype Message.offered_cops_msg

  datatype bribe_result = Push of pid * nid
                        | BribeChoose of bid list
                        (* Use if robber said nobribe in previous msg. *)
                        | BribeNone

  type robber_move_msg = {move : nid,
                          result : bribe_result}

  type inform_msg = (pid * nid * ptype * time * int) list

  type plan_msg = (pid * nid * ptype * time) list

  type vote_msg = bid list

  type vote_res_msg = bid option

  type from_inform_msg = (pid * inform_msg) list
  type from_plan_msg = (bid * plan_msg) list
	               
  type state = exn

  exception Super of string

  type graph = Graph.graph

  (* Data from the original world skeleton message. *)
  val get_name : super -> pid
  val get_robber : super -> pid
  val get_cops : super -> (pid * pid * pid * pid * pid)
  val get_cops_as_list : super -> pid list
  val get_nodes : super -> node list
  val get_banks_as_list : super -> node list
  val get_edges : super -> edge list
  val get_graph : super -> graph

  (* Data from the most recent world_msg. *)
  val get_time : super -> time
  val get_loot : super -> int
  (* XXX bank_vals, evidence, etc *)
  val get_dirty_cops : super -> bid list
  val get_controlling : super -> (bid * bid) list
  val get_smell : super -> int option
  val get_evidence : super -> (nid * time) list
  val get_player_locs : super -> (pid * nid * ptype) list
  val get_cop_locs : super -> (pid * nid * ptype) list
  val get_player_loc : super * pid -> nid * ptype
  val get_my_loc : super -> nid * ptype

  (* ... from the most recent from_inform_msg. *)
  val get_informs : super -> from_inform_msg
  val get_informs_by_informer : super * pid -> inform_msg

  (* ... from the most recent from_plan_msg. *)
  val get_plans : super -> from_plan_msg
  val get_plans_by_planner : super * pid -> plan_msg
  
  type dirtycop = {
    inform : super * world_msg * state -> inform_msg * state,
    plan : super * from_inform_msg * state -> plan_msg * state,
    vote : super * from_plan_msg * state -> vote_msg * state,

    move : super * vote_res_msg * state -> dirty_cop_move_msg * state,

    (* For forming dastardly plans! *)
    r_inform : super * world_msg * state -> inform_msg * state,
    r_plan : super * from_inform_msg * state -> plan_msg * state,
    r_vote : super * from_plan_msg * state -> vote_msg * state,
    r_results : super * vote_res_msg * state -> state
  }

  type cleancop = {
    register : unit -> register_msg * state,
    world_skeleton : super * world_skeleton_msg * state -> state,

    dirtify : super * state -> dirtycop * state,

    inform : super * world_msg * state -> inform_msg * state,
    plan : super * from_inform_msg * state -> plan_msg * state,
    vote : super * from_plan_msg * state -> vote_msg * state,

    move : super * vote_res_msg * state -> clean_cop_move_msg * state
  }

  type robber = {
    register : unit -> register_msg * state,
    world_skeleton : super * world_skeleton_msg * state -> state,

    inform : super * world_msg * state -> inform_msg * state,
    plan : super * from_inform_msg * state -> plan_msg * state,
    vote : super * from_plan_msg * state -> vote_msg * state,
    bribe : super * vote_res_msg * state -> bribe_msg * state,

    move : super * offered_cops_msg * state -> robber_move_msg * state
  }

  val BOTNAME : string ref

  val super_cop : cleancop -> Message.cleancop

  val super_robber : robber -> Message.robber
end
