(* regressions for graph. passes     
   Sat Jul  9 13:19:32 EST 2005 *)

structure T = Types
structure G = Graph  


fun try f x = f x
    handle (e as G.Graph s) => (print ("Graph error: " ^ s ^ "\n");
                                raise e)
         (* more handlers here... *)
         | (e as Fail s) => (print ("general fail: " ^ s ^ "\n");
                             raise e)

(*

 this is the graph represented below.

 ----> foot edge
 ====> car edge

                                    +---+
    A <------------ D <-------------| E | robberstart
    ,               /\              +---+
    |               ||                ^
    |               ||                |
    |               ||                |
    v             +---+               '
    B ----------> | C | ------------> F
                  +---+
                   hq

*)

val g = try G.tograph ([("a", T.Ordinary, 0, 0),
                        ("b", T.Ordinary, 0, 1),
                        ("c", T.Hq, 1, 0),
                        ("d", T.Ordinary, 1, 1),
                        ("e", T.RobberStart, 2, 2),
                        ("f", T.Ordinary, 0, 3)],

                       [("a", "b", T.Foot),
                        ("a", "c", T.Foot),
                        ("a", "f", T.Foot),
                        ("c", "d", T.Car),
                        ("d", "e", T.Foot),
                        ("e", "d", T.Foot),
                        ("f", "e", T.Foot),
                        ("d", "a", T.Foot),
                        ("b", "c", T.Foot),
                        ("c", "f", T.Foot)])

val ntn = G.name_to_nid g

fun p s = (print s; print "\n")
fun ts f (x::x'::xs) = (f x) ^ ", " ^ (ts f (x'::xs))
  | ts f [x] = (f x) ^ "\n"
  | ts f nil = "\n"

fun plist l = p (StringUtil.delimit ", " l)

fun plist' l = p (ts (fn (nd, pt) => 
           "(" ^ (G.nid_to_name g nd) ^ ", " ^ (Types.pttos pt) ^ ")") l)

fun plist'' (SOME(x)) = plist' x 
  | plist'' NONE = p "no path"

val plist = fn l => 
    let in
        (*    "should be: " *)
        print "           ";
        plist (ListUtil.sort String.compare 
               (map (G.nid_to_name g) l))
    end

fun test _ =
    (
     p "should be: a";
     plist (G.reachable g G.Forward (G.Only T.CopFoot) [ntn "a"] 0);

     p "should be: a, b, c, d, f";
     plist (G.reachable g G.Forward (G.Only T.CopFoot) [ntn "a"] 1);

     p "should be: a, b, c, f";
     plist (G.reachable g G.Forward (G.Only T.CopCar) [ntn "a"] 1);

     p "should be: a, b, c, f";
     plist (G.reachable g G.Forward (G.Only T.CopFoot) [ntn "c"] 1);

     p "should be: c, d, f";
     plist (G.reachable g G.Forward (G.Only T.CopCar) [ntn "c"] 1);

     p "should be: a, b, c, f";
     plist (G.reachable g G.Backward (G.Only T.CopFoot) [ntn "c"] 1);

     p "should be: a, b, c";
     plist (G.reachable g G.Backward (G.Only T.CopCar) [ntn "c"] 1);

     p "should be: a, c, d, e, f";
     plist (G.reachable g G.Forward (G.Only T.CopCar) [ntn "c"] 2);

     p "should be: a, b, c, d, e, f";
     plist (G.reachable g G.Forward (G.Only T.CopFoot) [ntn "a"] 2);

     p "should be: a, b, c, d, e, f";
     plist (G.reachable g G.Forward (G.Any T.CopFoot) [ntn "a"] 2);

     p "should be: a, b, c, d, e, f";
     plist (G.reachable g G.Forward (G.Any T.CopFoot) [ntn "f"] 3);

     p "should be (set): (b, foot) (c, foot), (d, foot), (f, foot)";
     plist' (G.adjacent g G.Forward (G.Any T.CopFoot) (ntn "a"));

     p "should be path: c";
     plist'' (G.shortest_path g (G.Only T.CopFoot) (ntn "f") (ntn "c"));

     p "should be path: f e";
     plist'' (G.shortest_path g (G.Only T.CopCar) (ntn "a") (ntn "e"));

     (* stay in car *)
     p "should be path: d a";
     plist'' (G.shortest_path g (G.Only T.CopCar) (ntn "c") (ntn "a"));

     (* get out of car *)
     p "should be path: a";
     plist'' (G.shortest_path g (G.Any T.CopCar) (ntn "c") (ntn "a"));
     
     p "should be path: a d";
     plist'' (G.shortest_path g (G.Only T.CopFoot) (ntn "c") (ntn "d"));

     ())

val _ = try test ()
