
(* The transformer (and related structures) allows bots to invoke
   the game logic to explore the outcomes of different moves. 
   
   Because the game logic depends on more about the state than is
   represented in the 'world' type (see protocol), we need a new
   type 'gamestate'. From it we can extract the world, or the world
   skeleton. *)

(* XXX twist-TODO: add the possibility pushing/bribing to the move
   set? *)
structure Transform =
struct

    exception Transform of string

    (* locations of the cops. bool: in a car? *)
    type exactcops = (Message.pid * Graph.nid * bool) list

    type exactrobber = Graph.nid

    (* PERF use vectors to store this stuff *)
    datatype ('cops, 'robber) gamestate =
        GS of { 
                (* forget the other stuff, it is useless for planning. *)
                skel : Graph.graph,

                (* turn number *)
                time : int,

                (* how much loot the robber has. *)
                loot : int,
                
                (* how much heat we've encountered *)
                temp : int,

                bank_vals : (Graph.nid * int) list,
                
                (* this is evidence for every node, tagged with
                   the time at which it was placed. known by
                   robber but not cops
                   *)
                evidence : (Graph.nid * int) list,
                
                cops : 'cops,
                
                (* location of robber *)
                robber : 'robber,
                
                (* bank transfers in progress. the int is the
                   scheduled turn before which it will happen.
                   (these are always even-numbered)
                   
                   THERE MAY BE MORE THAN ONE PER BANK.
                   however, an invariant is that there is
                   at most one bank per time.
                   
                   always exactly known by both *)
                transfers : (Graph.nid * int) list }

end